/********************************************************************
*
* Flash ROM writing code
*
* Andy Rabagliati   .   andyr@wizzy.com
* <URL http://www.wizzy.com/andyr/>
*
* Bits taken from spcecdt@armory.com (John DuBois) ihex2bin code,
*   and other places.
*
********************************************************************/

 #include <stdio.h>
 #include <string.h>
 #include <process.h>
 #include <stdlib.h>
 
 #define	FLASH_ADDR	0x7FF80000
 #define FLASH_SIZE	0x80000

 #define MAXLINE 1024

/* Globals */

char           *Filename = 0;	/* File being read, for use by error messages */
int             NR;		/* Number of line being processed, for use by
				 * error messages */
int             Trans[256];	/* Hex char to integer value lookup table */
char           *Name;
int             Dots = 0;
int             Verbose = 0;

 #define ISA             0x500

void
writeV3word(int offset, unsigned int data)
{
  __asm {
    ld              data;
    ld              offset;
    devsw;
  }
}
unsigned long
readV3word(int offset)
{
  unsigned long   data;
  __asm {
    ld              offset;
    devlw;
    st              data;
  }
  return (data);
}

void
led(int led)
{
  int             bits[] =
  {0x80, 0x02, 0x04, 0x10, 0x20, 0x40, 0x08, 0x01};
  static int             current = 0xff;

  if (led < 8) {
    current = current & (~bits[led]);
    writeV3word(ISA, 6);
    writeV3word(ISA + 4, current);
  } else {
    led -= 8;
    current = current | (bits[led]);
    writeV3word(ISA, 6);
    writeV3word(ISA + 4, current);
  }
}

/************************************************************************
*
*  Module:	flash.c
*
*  Description:	Flash erasure/modification routines
*
************************************************************************/

/************************************************************************/
/* */
/* Table of all the supported FLASH devices			 */
/* */
/************************************************************************/

typedef struct {
  char            ID1, ID2;
  int             Capacity;
  char           *Description;
}               Flash_ID_t;

Flash_ID_t      Flash_IDs[] = {{0x01, 0x20, 128 * 1024, "AMD Am29F010"},
{0x01, 0xA4, 512 * 1024, "AMD Am29F040"}};
int             Num_Flash_IDs = 2;

/************************************************************************
*
*  Subroutine:	ID_Flash
*
*  Prototype:	int	ID_Flash(void)
*
*  Description:	Identify the flash memory present in the system.  If it
+		is one that is supported (in the ID table), return the
+		ID structure number associated with the flash.
*
*  Inputs:
+		None
*
*  Return value:
+		>= 0	Number of the structure in the ID table which
+			describes the FLASH in the system
+		< 0	Not one of the supported devices or could not
+			read ID
*
************************************************************************/

int
ID_Flash(void)
{
  volatile char  *Fptr;
  char            ID1, ID2;
  int             i;

  Fptr = (char *) FLASH_ADDR;

  /* Execute Autoselect Command */

  Fptr[0x5555] = 0xAA;
  ProcWait(100);
  Fptr[0x2AAA] = 0x55;
  ProcWait(100);
  Fptr[0x5555] = 0x90;
  ProcWait(100);

  /* Read ID bytes */

  ID1 = Fptr[0];
  ProcWait(100);
  ID2 = Fptr[1];
  ProcWait(100);

  /* Execute Read/Reset Command */

  Fptr[0x5555] = 0xAA;
  ProcWait(100);
  Fptr[0x2AAA] = 0x55;
  ProcWait(100);
  Fptr[0x5555] = 0xF0;
  ProcWait(100);

  for (i = 0; i < Num_Flash_IDs; i++)
    if (Flash_IDs[i].ID1 == ID1 && Flash_IDs[i].ID2 == ID2)
      break;

  if (i < Num_Flash_IDs && Flash_IDs[i].ID1 == ID1 && Flash_IDs[i].ID2 == ID2)
    return (i);
  else
    return (-1);
}

/************************************************************************
*
*  Subroutine:	Erase_Flash
*
*  Prototype:	int	Erase_Flash(void)
*
*  Description:	Erase the FLASH memory installed in the system.
*
*  Inputs:
+		None
*
*  Return value:
+		0	Successful
+		!= 0	Unsuccessful
*
************************************************************************/

int
Erase_Flash(void)
{
  char            TestByte;
  volatile char  *Fptr;

  Fptr = (char *) FLASH_ADDR;

  /* Erase flash */

  Fptr[0x5555] = 0xAA;		/* Output 0xAA to 0x5555 */
  Fptr[0x2AAA] = 0x55;		/* Output 0x55 to 0x2AAA */
  Fptr[0x5555] = 0x80;		/* Output 0x80 to 0x5555 */
  Fptr[0x5555] = 0xAA;		/* Output 0xAA to 0x5555 */
  Fptr[0x2AAA] = 0x55;		/* Output 0x55 to 0x2AAA */
  Fptr[0x5555] = 0x10;		/* Output 0x10 to 0x5555 */

  /* Wait for Embedded Erase Algorithm to complete. */
  while ((TestByte = *Fptr) != 0xFF)
    if ((TestByte & 0x20) != 0) {
      return -1;
    }
  return (0);
}

void
dump(char *p, int length)
{
  while (length > 0) {
    int             i;
    printf("\n%8x:", (int) p);
    for (i = 0; i < 16; i++) {
      printf(" %02x", *p++);
    }
    length -= 16;
  }
  printf("\n");
}

/************************************************************************
*
*  Subroutine:	Program_Flash
*
*  Prototype:	int	Program_Flash(char *ImageData, int offset,
*                                                   int ImageBytes)
*
*  Description:	Reprogram the FLASH memory installed in the system,
+		given a pointer to the data and the number of bytes.
*
*  Inputs:
+		ImageData	Pointer to the data to be programmed
+		ImageBytes	Number of bytes pointed to by ImageData
*
*  Return value:
+		0	FLASH successfully reprogrammed
+		!= 0	Failure during FLASH reprogramming
*
************************************************************************/

int
Program_Flash(unsigned char *ImageData, long Offset, int ImageBytes)
{
  volatile char  *Fptr;
  char            TestByte;
  int             CurByte;

  Fptr = (char *) FLASH_ADDR;

  /* Program the flash */

  for (CurByte = 0; CurByte < ImageBytes; CurByte++) {
    if (Fptr[Offset + CurByte] != 0xff)
      return 1;			/* cannot program location */
    if (ImageData[CurByte] != 0xFF) {
      /* Output Byte Program command */

      Fptr[0x5555] = 0xAA;	/* Output 0xAA to 0x5555 */
      Fptr[0x2AAA] = 0x55;	/* Output 0x55 to 0x2AAA */
      Fptr[0x5555] = 0xA0;	/* Output 0xA0 to 0x5555 */
      Fptr[Offset + CurByte] = ImageData[CurByte];

      /* Wait for byte programming to complete */
      while ((TestByte = Fptr[Offset + CurByte]) != ImageData[CurByte]);
    }
  }
  return (0);
}

/*
 * Make Arr be a hex char to integer lookup table, with the indexes of
 * invalid characters set to -1.
 */
void
MakeHexTrans(Arr)
  int             Arr[];
{
  int             i;

  for (i = 0; i < 256; i++)
    Arr[i] = -1;
  for (i = '0'; i <= '9'; i++)
    Arr[i] = i - '0';
  for (i = 'a'; i <= 'f'; i++)
    Arr[i] = i - 'a' + 10;
  for (i = 'A'; i <= 'F'; i++)
    Arr[i] = i - 'A' + 10;
}

 #define h2b(p) ((p+=2),((Trans[p[-2]] << 4) + Trans[p[-1]]))
 #define BadRecordMsg(Name,NR,Filename,S) \
fprintf(stderr, "%s%s: Bad record on line %d of file \"%s\":\n%s\n",\
(Dots ? "\n" : ""),(Name),(NR),(Filename),(S))


/*
 * Format of an Intel hex record: :nnaaaa00dd....ddss nn: Number of data
 * bytes aaaa: Address dd: data bytes ss: 0-sum corrective checksum for all
 * bytes on line.
 */
/*
 * ihex2bin: convert an Intel-hex format record to binary data. Input
 * variables: In is the input record. Output variables: Out is the output
 * data. *OffsetP is set to the load address given in the record. Return
 * value: On successful conversion, the number of bytes in the record is
 * returned. On failure, -1 is returned.
 */
int
ihex2bin(char *In, unsigned char Out[], long *OffsetP)
{
  int             i, j;
  int             Val;		/* Hex value of a char */
  unsigned        Sums[2];	/* Checksums of odd & even chars */
  unsigned        Checksum;	/* Checksum of the record */
  int             NumChars;	/* Number of data byte characters found */
  unsigned        NumSpeced;	/* Number of data byte characters expected */
  int             NumBytes;	/* Number of bytes in record */
  char           *InP = In;	/* Current data being processed */
  int             ByteNum;	/* Byte being converted */
  int             Skip;
  /*
   * char            ConvBuf[MAXLINE * 2];	Buffer for NoCtlCpy()
   * conversion
   */

  if (In[0] == ':')
    Skip = 0;
  else {
    if (!Skip) {
      printf("%s\n", In);
      fputs("^ Did not start with ':'.  Ignored.\n", stderr);
    }
    return -1;
  }
  InP = In + 7;
  switch (h2b(InP)) {
  case 0:
    break;
  case 1:
    return -2;			/* End of data in this file */
    break;
  case 2:			/* Extended offset */
    *OffsetP = h2b(InP) << 8;
    *OffsetP += h2b(InP);
    return -3;

  default:
    printf("%s\n", In);
    fprintf(stderr, "       ^^ Unknown record type %c%c.  Ignored.\n",
	    In[7], In[8]);
    return -1;
  }
  /*
   * Make sure record contains only hex characters (after first char), and
   * check checksum
   */
  Sums[0] = Sums[1] = 0;
  for (i = 1; (Val = Trans[In[i]]) >= 0; i++)
    Sums[i & 1] += Val;
  if (In[i]) {
    printf("%s\n", In);
    for (j = 0; j < i; j++)
      fputc(' ', stderr);
    fprintf(stderr, "^  is not a hex char.  Record ignored.\n");
    return -1;
  }
  Checksum = ((Sums[1] << 4) + Sums[0]) & 0xFF;
  if (Checksum) {
    printf("%s\n", In);
    fprintf(stderr,
    "Checksum = %u; should be zero.  Record processed anyway.\n", Checksum);
  }
  /* 00 records contain 11 non-data characters */
  NumChars = i - 11;
  InP = In + 1;			/* Skip ':' */
  if ((NumSpeced = h2b(InP)) << 1 != NumChars) {
    printf("%s\n", In);
    fprintf(stderr,
	    "Record specifies %d data bytes (%d characters); %d data characters found.\n" \
	 "Record processed anyway.\n", NumSpeced, NumSpeced << 1, NumChars);
  }
  *OffsetP = h2b(InP) << 8;
  *OffsetP += h2b(InP);
  InP += 2;			/* Skip record type */
  NumBytes = NumChars >> 1;
  for (ByteNum = 0; ByteNum < NumBytes; ByteNum++)
    Out[ByteNum] = h2b(InP);
  return NumBytes;
}

int
ProcFile(InFile)
  FILE           *InFile;
{
  char            Buf[MAXLINE + 1];	/* Input record buffer */
  unsigned char   bBuf[MAXLINE / 2];	/* Output buffer */
  long            Offset;	/* File offset to write record at */
  long            Extended = 0;	/* Segment address */
  int             nbytes;	/* Number of bytes in record */
  int             Len;		/* Length of input line */
  int             Written;	/* Number of bytes actually written */
  int             Tot = 0;	/* Total number of bytes written */

  while (fgets(Buf, MAXLINE + 1, InFile)) {
    NR++;
    if (Buf[(Len = strlen(Buf)) - 1] == '\n')
      Buf[Len - 1] = '\0';
    else {
      if (Dots)
	fputc('\n', stderr);
      fprintf(stderr, "Line %d of file %s is too long; max line is %d.\n",
	      NR, Filename, MAXLINE);
      exit(5);
    }
    switch (nbytes = ihex2bin(Buf, bBuf, &Offset)) {
    case 0:
    case -1:
      continue;
      break;
    case -2:
      return Tot;;
    case -3:
      Extended = Offset << 4;
      if (Verbose)
	fprintf(stderr, "Extended offset set to %04lX.\n", Extended);
      if (Dots)
	fprintf(stderr, "\n<%5lX> ", Extended);
      continue;
      break;
    }
    if (Verbose)
      fprintf(stderr, "Writing %2d byte(s) at %04lX.\n", nbytes,
	      Extended + Offset);
    switch ((Extended + Offset) & 0x600) {
      case 0x000:
        led(0);
	break;
      case 0x200:
        led(1);
	break;
      case 0x400:
        led(8);
	break;
      case 0x600:
        led(9);
	break;
    }
    if (Dots) {
      if (!((Extended + Offset) & 0x7ff))
	fprintf(stderr, "\n<%5lX> ", (Extended + Offset));
      else
	fprintf(stderr, ".");
      fflush(stderr);
    }
    if (Program_Flash(bBuf, Extended + Offset, nbytes)) {
      if (Dots)
	fputc('\n', stderr);
      fprintf(stderr,
	      "%s: Tried to write %d bytes at %lx; only %d written.\n",
	      Name, nbytes, Extended + Offset, Written);
      exit(6);
    }
    Tot += nbytes;
  }
  led(8);
  led(9);
  if (Dots)
    fprintf(stderr, "\n");
  return Tot;
}

int
checksum()
{
  int i, csum = 0;
  volatile char  *Fptr;

  Fptr = (char *) FLASH_ADDR;
  for (i=0; i<FLASH_SIZE; i++)
    csum += *Fptr++;
  return csum;
}

int
main(int argc, char *argv[])
{
  int             i, whatflash;
  FILE           *InFile;
  int             all = 0;
  int             Tot = 0;	/* Total number of bytes written */

  Name = argv[0];
  i = 1;
  while (i < argc) {
    char           *c;
    c = argv[i];
    if (*c == '-') {
      c++;
      switch (*c) {
      case 'a':
	all = 1;
	break;
      case 'd':
	Dots = 1;
	break;
      case 'v':
	Verbose = 1;
	break;
      }
    } else {
      Filename = c;
    }
    i++;
  }
  if (!Filename) {
    printf("Usage: %s filename.hex\n", Name);
    exit(1);
  }
  NR = 0;
  if (Verbose)
    printf("Flashing from %s\n", Filename);
  MakeHexTrans(Trans);
  whatflash = ID_Flash();
  if (whatflash < 0) {
    printf("Cannot determine flash type\n");
    exit(1);
  }
  if (Verbose)
    printf("Flash identified as %dK %s\n",
	   Flash_IDs[whatflash].Capacity / 1024,
	   Flash_IDs[whatflash].Description);

  led(8);				/* leds off */
  writeV3word(ISA, 0);
  if (!all)
    while (readV3word(ISA + 4) & 0x20)
       ProcWait(1000);			/* look for jumper */
  if (!(InFile = fopen(Filename, "r"))) {
    fprintf(stderr, "%s: Could not open input file \"%s\": ",
	    Name, Filename);
    perror("");
    exit(2);
  }
  if (Verbose)
    printf("File %s opened\n", Filename);
  led(0);
  led(1);
  if (Dots)
    printf("Erasing %dK %s\n",
	   Flash_IDs[whatflash].Capacity / 1024,
	   Flash_IDs[whatflash].Description);
  if (Erase_Flash()) {
    printf("Failed to erase flash\n");
    exit(2);
  }
  if (Verbose)
    printf("Flash erased successfully\n");
  Tot += ProcFile(InFile);
  fclose(InFile);
  writeV3word(ISA, 0);
  i = readV3word(ISA + 4) & 3;	/* board ID */
  fprintf(stderr,
    "\nBrick %d successfully programmed from %s - Checksum = %04X\n",
	  i, Filename, checksum() & 0xFFFF);
  led(0);
  ProcWait(15625*2);
}
