/*
*   ihex2bin: convert Intel hex data to binary file.
*   93/03/08 john h. dubois iii (john@armory.com)
*/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>

int ProcFile(FILE *InFile);
void MakeHexTrans(int Arr[]);

#define MAXLINE 1024

char *FILENAME;	/* File being read, for use by error messages */
int NR;		/* Number of line being processed, for use by error messages */
int Trans[256];	/* Hex char to integer value lookup table */
char *Name;	/* Name of this process, for use by help & error messages */
int verbose = 0;	/* Whether to print running processing info */
int Dots = 0;		/* Whether to print a dot for every record read */
int Skip = 0;		/* Skip initial lines that do not begin with ':' */
char *Comment = "@(#) ihex2bin 1.1 96/07/03";

main(argc,argv)
int argc;
char *argv[];
{
    extern char *optarg;
    extern int optind;
    int Force = 0;	/* Use output file even if it exists */
    int Quiet = 0;	/* Do not print status messages */
    char c;
    FILE *InFile;
    char *OutFile = NULL;
    char *Usage = "Usage: %s [-dhqsv] [-[cou] output-file] [input-file ...]\n";
    struct stat statbuf;
    char *OpenFlags = "w";	/* Flags to pass to fopen */
    int Tot = 0;	/* Total number of bytes written */

    if (Name = strrchr(argv[0],'/'))
	Name++;
    else
	Name = argv[0];
    while ((c = getopt(argc, argv, "dhqsvc:o:u:")) != -1)
	switch (c) {
	case 'd':
	    Dots = 1;
	    break;
	case 'u':
	    OpenFlags = "r+";
	case 'o':
	    Force = 1;
	case 'c':
	    if (OutFile) {
		fprintf(stderr,
		"%s: Cannot give more than one of c, o, or u options.\n",Name);
		fprintf(stderr,Usage,Name);
		exit(1);
	    }
	    OutFile = optarg;
	    break;
	case 'v':
	    verbose = 1;
	    break;
	case 'q':
	    Quiet = 1;
	    break;
	case 's':
	    Skip = 1;
	    break;
	case 'h':
	    printf(
	    "%s: convert Intel-hex format data to a binary file.\n",Name);
	    printf(Usage,Name);
	    fputs(
"-o: Overwrite or create output-file.  Output-file is truncated if it exists.\n"\
"-u: Update or create output-file.  Output-file is not truncated if it exists;\n"\
"    further records are written into it, possibly overwriting parts of it.\n"\
"-c: Check whether output-file exists and create it if it does not.\n"\
"    If it does, an error message is printed and it is not touched.\n"\
"-d: Print a dot on stderr for every record written.\n"\
"-q: Quiet operation: no status messages are printed.\n"\
"-s: Initial lines that do not begin with ':' are skipped without warning.\n"\
"-h: Print this help information.\n"\
"-v: Print a brief description of records as they are written.\n"\
"\n"\
"If output-file is not given or is given as '-',\n"\
"output is written to the standard output.\n"\
"If records do not start at address 0 or contain any non-contiguous records,\n"\
"output must be a real file, not a pipe or tty (to allow seeking).\n"\
"If no input-file is given, input is read from the standard input.\n"\
,stdout);
	    exit(0);
	case '?':
	    fprintf(stderr,"%s: Invalid flag: '%c'.\n",Name,c);
	    fprintf(stderr,Usage,Name);
	    exit(1);
	}
    if (OutFile && !(*OutFile == '-' && !OutFile[1])) {
	if (!Force && !stat(OutFile,&statbuf) && S_ISREG(statbuf.st_mode)) {
	    fprintf(stderr,
	    "%s: %s: file exists.  Use -o or -u to write to it.\n",
	    Name,OutFile);
	    exit(4);
	}
	if (!freopen(OutFile,OpenFlags,stdout)) {
	    fprintf(stderr,"%s: Could not open output file \"%s\": ",
	    Name,OutFile);
	    perror("");
	    exit(2);
	}
    }
    MakeHexTrans(Trans);
    if (optind < argc)
	for (; optind < argc; optind++) {
	    if (!(InFile = fopen(argv[optind],"r"))) {
		if (Dots)
		    fputc('\n',stderr);
		fprintf(stderr,"%s: Could not open input file \"%s\": ",
		Name,argv[optind]);
		perror("");
		exit(2);
	    }
	    NR = 0;
	    FILENAME = argv[optind];
	    Tot += ProcFile(InFile);
	    fclose(InFile);
	}
    else {
	NR = 0;
	FILENAME = "stdin";
	Tot = ProcFile(stdin);
    }
    if (!Quiet) {
	if (Dots)
	    fputc('\n',stderr);
	fprintf(stderr,"%d bytes written.\n",Tot);
    }
}

/*
*   Make Arr be a hex char to integer lookup table,
*   with the indexes of invalid characters set to -1.
*/
void MakeHexTrans(Arr)
int Arr[];
{
    int i;

    for (i = 0; i < 256; i++)
	Arr[i] = -1;
    for (i = '0'; i <= '9'; i++)
	Arr[i] = i - '0';
    for (i = 'a'; i <= 'f'; i++)
	Arr[i] = i - 'a' + 10;
    for (i = 'A'; i <= 'F'; i++)
	Arr[i] = i - 'A' + 10;
}

#define h2b(p) ((p+=2),((Trans[p[-2]] << 4) + Trans[p[-1]]))
#define BadRecordMsg(Name,NR,Filename,S) \
fprintf(stderr, "%s%s: Bad record on line %d of file \"%s\":\n%s\n",\
(Dots ? "\n" : ""),(Name),(NR),(Filename),(S))

int ProcFile(InFile)
FILE *InFile;
{
    char Buf[MAXLINE+1];	/* Input record buffer */
    char bBuf[MAXLINE / 2];	/* Output buffer */
    long Offset;		/* File offset to write record at */
    int nbytes;			/* Number of bytes in record */
    int Len;			/* Length of input line */
    int Written;		/* Number of bytes actually written */
    static int CurOffset = 0;	/* Current file offset */
    int Tot = 0;		/* Total number of bytes written */

    while (fgets(Buf,MAXLINE+1,InFile)) {
	NR++;
	if (Buf[(Len = strlen(Buf)) - 1] == '\n')
	    Buf[Len - 1] = '\0';
	else {
	    if (Dots)
		fputc('\n',stderr);
	    fprintf(stderr,"Line %d of file %s is too long; max line is %d.\n",
	    NR,FILENAME,MAXLINE);
	    exit(5);
	}
	switch (nbytes = ihex2bin(Buf,bBuf,&Offset)) {
	case -1:
	    continue;
	    break;
	case -2:
	    return Tot;;
	}
	if (Offset != CurOffset && fseek(stdout,Offset,SEEK_SET)) {
	    BadRecordMsg(Name,NR,FILENAME,Buf);
	    fprintf(stderr, "Cannot seek on output from %d to %d.\n",
	    CurOffset,Offset);
	    exit(3);
	}
	if (verbose)
	    fprintf(stderr,"Writing %2d byte(s) at %04X.\n",nbytes,Offset);
	if (Dots)
	    write(2,".",1);
	if ((Written = fwrite(bBuf,1,nbytes,stdout)) != nbytes) {
	    if (Dots)
		fputc('\n',stderr);
	    fprintf(stderr, 
	    "%s: Tried to write %d bytes at %d; only %d written.\n",
	    Name,nbytes,Offset,Written);
	    exit(6);
	}
	/* Keep track of offset instead of using ftell() because
	   ftell() will fail on a pipe */
	CurOffset = Offset + nbytes;
	Tot += nbytes;
    }
    return Tot;
}

/*
*   Format of an Intel hex record:
*   :nnaaaa00dd....ddss
*   nn: Number of data bytes
*   aaaa: Address
*   dd: data bytes
*   ss: 0-sum corrective checksum for all bytes on line.
*/
/*
*   ihex2bin: convert an Intel-hex format record to binary data.
*   Input variables: In is the input record.
*   Output variables:
*   Out is the output data.
*   *OffsetP is set to the load address given in the record.
*   Return value:
*   On successful conversion, the number of bytes in the record is returned.
*   On failure, -1 is returned.
*/
ihex2bin(In,Out,OffsetP)
char *In;
unsigned char Out[];
long *OffsetP;
{
    int i,j;
    int Val;		/* Hex value of a char */
    unsigned Sums[2];	/* Checksums of odd & even chars */
    unsigned Checksum;	/* Checksum of the record */
    int NumChars;	/* Number of data byte characters found */
    unsigned NumSpeced;	/* Number of data byte characters expected */
    int NumBytes;	/* Number of bytes in record */
    char *InP = In;	/* Current data being processed */
    int ByteNum;	/* Byte being converted */
    char ConvBuf[MAXLINE * 2]; /* Buffer for NoCtlCpy() conversion */

    if (In[0] == ':')
	Skip = 0;
    else {
	if (!Skip) {
	    BadRecordMsg(Name,NR,FILENAME,In);
	    fputs("^ Did not start with ':'.  Ignored.\n",stderr);
	}
	return -1;
    }
    InP = In + 7;
    switch (h2b(InP)) {
    case 0:
	break;
    case 1:
	return -2;	/* End of data in this file */
	break;
    default:
	BadRecordMsg(Name,NR,FILENAME,In);
	fprintf(stderr,"       ^^ Unknown record type %c%c.  Ignored.\n",
	In[7],In[8]);
	return -1;
    }
    /* Make sure record contains only hex characters (after first char),
       and check checksum */
    Sums[0] = Sums[1] = 0;
    for (i = 1; (Val = Trans[In[i]]) >= 0; i++)
	Sums[i & 1] += Val;
    if (In[i]) {
	BadRecordMsg(Name,NR,FILENAME, NoCtlCpy(ConvBuf,In,MAXLINE * 2,-2));
	for (j = 0; j < i; j++)
	    fputc(' ',stderr);
	fprintf(stderr,"^ '%s' is not a hex char.  Record ignored.\n",
	NoCtlCpy(ConvBuf,In + i,1,-2));
	return -1;
    }
    Checksum = ((Sums[1] << 4) + Sums[0]) & 0xFF;
    if (Checksum) {
	BadRecordMsg(Name,NR,FILENAME,In);
	fprintf(stderr,
	"Checksum = %u; should be zero.  Record processed anyway.\n",Checksum);
    }
    /* 00 records contain 11 non-data characters */
    NumChars = i - 11;
    InP = In + 1;	/* Skip ':' */
    if ((NumSpeced = h2b(InP)) << 1 != NumChars) {
	BadRecordMsg(Name,NR,FILENAME,In);
	fprintf(stderr, 
"Record specifies %d data bytes (%d characters); %d data characters found.\n"\
	"Record processed anyway.\n", NumSpeced,NumSpeced << 1,NumChars);
    }
    *OffsetP = h2b(InP) << 8;
    *OffsetP += h2b(InP);
    InP += 2; 	/* Skip record type */
    NumBytes = NumChars >> 1;
    for (ByteNum = 0; ByteNum < NumBytes; ByteNum++)
	Out[ByteNum] = h2b(InP);
    return NumBytes;
}
